"""
Just a regular `setup.py` file.

Author: Nikolay Lysenko
"""


import os
from setuptools import setup, find_packages


current_dir = os.path.abspath(os.path.dirname(__file__))

description = 'Algorithmic composition of twelve-tone fragments.'
with open(os.path.join(current_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='dodecaphony',
    version='0.1.5',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Nikolay-Lysenko/dodecaphony',
    author='Nikolay Lysenko',
    author_email='nikolay-lysenco@yandex.ru',
    license='MIT',
    keywords=[
        'ai_music',
        'algorithmic_composition',
        'dodecaphony',
        'generative_art',
        'twelve_tone',
    ],
    packages=find_packages(exclude=["tests"]),
    package_data={
        'dodecaphony': [
            'configs/default_config.yml',
            'configs/sinethesizer_presets.yml'
        ]
    },
    python_requires='>=3.9',
    install_requires=[
        'pretty-midi',
        'PyYAML',
        'sinethesizer>=0.6,<0.7',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Artistic Software',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ]
)



"""
Run specified tasks.

Author: Nikolay Lysenko
"""


import argparse
import os
from pkg_resources import resource_filename

import yaml

from .evaluation import evaluate, parse_scoring_sets_registry
from .fragment import FragmentParams, initialize_fragment
from .optimization import optimize_with_local_search
from .rendering import render
from .transformations import create_transformations_registry


def parse_cli_args() -> argparse.Namespace:
    """
    Parse arguments passed via Command Line Interface (CLI).

    :return:
        namespace with arguments
    """
    parser = argparse.ArgumentParser(description='Algorithmic composition of dodecaphonic music.')
    parser.add_argument(
        '-c', '--config_path', type=str, default=None, help='path to configuration file'
    )
    parser.add_argument(
        '-n', '--n_fragments', type=int, default=1, help='number of fragments to render'
    )
    cli_args = parser.parse_args()
    return cli_args


def main() -> None:
    """Parse CLI arguments and run requested tasks."""
    cli_args = parse_cli_args()

    default_config_path = 'configs/default_config.yml'
    default_config_path = resource_filename(__name__, default_config_path)
    config_path = cli_args.config_path or default_config_path
    with open(config_path) as config_file:
        settings = yaml.load(config_file, Loader=yaml.FullLoader)

    fragment_params = FragmentParams(**settings['fragment'])
    initial_fragment = initialize_fragment(fragment_params)

    scoring_sets_registry = parse_scoring_sets_registry(settings['scoring_sets'])
    scoring_sets = settings['evaluation']['scoring_sets']
    transformations_registry = create_transformations_registry(
        settings['optimization'].pop('max_rotation'),
        settings['optimization'].pop('max_transposition_in_semitones')
    )
    fragments = optimize_with_local_search(
        initial_fragment,
        transformation_registry=transformations_registry,
        scoring_sets=scoring_sets,
        scoring_sets_registry=scoring_sets_registry,
        **settings['optimization']
    )

    reports = []
    for fragment in fragments:
        _, report = evaluate(fragment, scoring_sets, scoring_sets_registry, report=True)
        reports.append(report)
    print("\nEvaluation of selected fragments:\n")
    print(*reports, sep="\n\n")
    print()

    results_dir = settings['rendering']['dir']
    if not os.path.isdir(results_dir):
        os.mkdir(results_dir)
    for fragment, report in zip(fragments[:cli_args.n_fragments], reports):
        rendering_params = settings['rendering']
        rendering_params['meta_information'] = (
            f"Config path: {os.path.abspath(config_path)}\n\n"
            f"Evaluation scores:\n{report}"
        )
        render(fragment, rendering_params)


if __name__ == '__main__':
    main()



"""
Evaluate a fragment.

Author: Nikolay Lysenko
"""


import itertools
import math
import re
import string
from collections import Counter
from functools import cache
from typing import Any, Callable, Optional

from .fragment import Event, Fragment
from .music_theory import (
    IntervalTypes,
    N_SEMITONES_PER_OCTAVE,
    get_mapping_from_pitch_class_to_diatonic_scales,
    get_type_of_interval,
)
from .utils import compute_rolling_aggregate


SCORING_SETS_REGISTRY_TYPE = dict[
    str,
    list[tuple[Callable[..., float], dict[float, float], dict[str, Any]]]
]


def evaluate_absence_of_aimless_fluctuations(
        fragment: Fragment, penalties: dict[int, float], window_size: int
) -> float:
    """
    Evaluate melodic fluency based on absence of aimless fluctuations within narrow ranges.

    :param fragment:
        a fragment to be evaluated
    :param penalties:
        mapping from range (in semitones) covered within a window to penalty applicable to ranges
        of not greater size; it is recommended to set maximum penalty to 1
    :param window_size:
        size of rolling window (in events)
    :return:
        multiplied by -1 penalty for narrowness averaged over all windows (including those that
        are not penalized)
    """
    numerator = 0
    denominator = 0
    for melodic_line in fragment.melodic_lines:
        pitches = [event.position_in_semitones for event in melodic_line]
        pitches = [x for x in pitches if x is not None]
        rolling_mins = compute_rolling_aggregate(pitches, min, window_size)[window_size-1:]
        rolling_maxs = compute_rolling_aggregate(pitches, max, window_size)[window_size-1:]
        borders = zip(rolling_mins, rolling_maxs)
        for lower_border, upper_border in borders:
            range_width = upper_border - lower_border
            raw_penalties = [v for k, v in penalties.items() if k >= range_width]
            penalty = max(raw_penalties) if raw_penalties else 0
            numerator -= penalty
            denominator += 1
    score = numerator / denominator
    return score


def evaluate_absence_of_doubled_pitch_classes(fragment: Fragment) -> float:
    """
    Evaluate absence of vertical intervals that are whole multipliers of octave.

    If the same pitch class sounds in two or more melodic lines simultaneously,
    harmonic stability increases significantly and it interrupts musical flow associated
    with the twelve-tone technique.

    :param fragment:
        a fragment to be evaluated
    :return:
        minus one multiplied by number of pairs of simultaneously sounding events
        of the same pitch class and divided by total number of sonorities
    """
    score = 0
    for sonority in fragment.sonorities:
        for first, second in itertools.combinations(sonority, 2):
            if first.pitch_class == 'pause' or second.pitch_class == 'pause':
                continue
            interval = first.position_in_semitones - second.position_in_semitones
            if interval % N_SEMITONES_PER_OCTAVE == 0:
                score -= 1
    score /= len(fragment.sonorities)
    return score


def evaluate_absence_of_simultaneous_skips(
        fragment: Fragment, min_skip_in_semitones: int = 5, max_skips_share: float = 0.65
) -> float:
    """
    Evaluate absence of simultaneous large enough skips.

    :param fragment:
        a fragment to be evaluated
    :param min_skip_in_semitones:
        minimum size (in semitones) of a melodic interval to be considered a large enough skip
    :param max_skips_share:
        maximum share of melodic lines with skips between adjacent sonorities not to be penalized
        for this pair of sonorities
    :return:
        minus one multiplied by fraction of sonorities with enough number of simultaneous skips
    """
    score = 0
    for first_sonority, second_sonority in zip(fragment.sonorities, fragment.sonorities[1:]):
        n_melodic_intervals = 0
        n_skips = 0
        for first, second in zip(first_sonority, second_sonority):
            if first.pitch_class == 'pause' or second.pitch_class == 'pause':
                continue
            n_melodic_intervals += 1
            interval_size = abs(first.position_in_semitones - second.position_in_semitones)
            n_skips += int(interval_size >= min_skip_in_semitones)
        if n_melodic_intervals > 0 and n_skips / n_melodic_intervals >= max_skips_share:
            score -= 1
    score /= len(fragment.sonorities) - 1
    return score


def evaluate_absence_of_voice_crossing(
        fragment: Fragment, n_semitones_to_penalty: dict[int, float]
) -> float:
    """
    Evaluate absence of voice crossing.

    Voice crossing may result in wrong perception of tone row and in incoherence of the fragment.

    :param fragment:
        a fragment to be evaluated
    :param n_semitones_to_penalty:
        mapping from size of vertical interval between a pair of voices (this size is assumed to be
        non-positive) to penalty for this interval
    :return:
        minus one multiplied by average over all vertical intervals penalty
    """
    numerator = 0
    denominator = 0
    for sonority in fragment.sonorities:
        for first, second in itertools.combinations(sonority, 2):
            if first.pitch_class == 'pause' or second.pitch_class == 'pause':
                continue
            interval = first.position_in_semitones - second.position_in_semitones
            if interval <= 0:
                numerator -= n_semitones_to_penalty.get(interval, 1)
            denominator += 1
    score = numerator / denominator
    return score


def evaluate_cadence_duration(
        fragment: Fragment,
        max_duration: float,
        last_sonority_weight: float,
        last_notes_weight: float
) -> float:
    """
    Evaluate that cadence has enough duration.

    :param fragment:
        a fragment to be evaluated
    :param max_duration:
        maximum enough duration (in reference beats); higher durations do not increase score
    :param last_sonority_weight:
        weight that determines contribution of last sonority duration to final score
    :param last_notes_weight:
        weight that determines contribution of duration of last note of each melodic line
        to final score; this term is rather a mean than a goal, because it rewards intermediate
        steps when last sonority duration is unchanged, but one of the last notes extends its
        duration
    :return:
        minus one multiplied by missing in last sonority and last notes fraction of `max_duration`
    """
    clipped_durations = [min(event.duration, max_duration) for event in fragment.sonorities[-1]]
    last_sonority_duration = min(clipped_durations)
    avg_last_note_duration = sum(clipped_durations) / len(clipped_durations)
    total_weight = last_sonority_weight + last_notes_weight
    last_sonority_weight /= total_weight
    last_notes_weight /= total_weight
    last_sonority_term = last_sonority_weight * (last_sonority_duration / max_duration - 1)
    last_notes_term = last_notes_weight * (avg_last_note_duration / max_duration - 1)
    score = last_sonority_term + last_notes_term
    return score


def evaluate_climax_explicity(
        fragment: Fragment,
        height_penalties: dict[int, float],
        duplication_penalty: float
) -> float:
    """
    Evaluate goal-orientedness of melodic lines by explicity of their climax points.

    :param fragment:
        a fragment to be evaluated
    :param height_penalties:
        mapping from interval size (in semitones) between average pitch of a line and climax of
        this line to penalty for such interval; values for missing keys are forward-propagated
    :param duplication_penalty:
        penalty for each non-first occurrence of line's highest pitch within this line
    :return:
        summed over melodic lines penalty for not so high climax points and for duplications
        of climax points within the same line
    """
    score = 0
    for melodic_line in fragment.melodic_lines:
        pitches = [event.position_in_semitones for event in melodic_line]
        pitches = [x for x in pitches if x is not None]

        climax_pitch = max(pitches)
        average_pitch = sum(pitches) / len(pitches)
        interval_size = climax_pitch - average_pitch
        raw_height_penalties = [v for k, v in height_penalties.items() if k >= interval_size]
        height_penalty = max(raw_height_penalties) if raw_height_penalties else 0
        score -= height_penalty

        n_duplications = len([x for x in pitches if x == climax_pitch]) - 1
        score -= duplication_penalty * n_duplications
    score /= len(fragment.melodic_lines)
    return score


def evaluate_consistency_of_rhythm_with_meter(
        fragment: Fragment, consistent_patterns: list[list[float]]
) -> float:
    """
    Evaluate ease of deriving meter from rhythms of melodic lines.

    :param fragment:
        a fragment to be evaluated
    :param consistent_patterns:
        list of options to split a measure time span into time spans of individual notes
        considered to be consistent with meter
    :return:
        minus one multiplied by number of measures which are split in inconsistent manner
    """
    score = 0
    for durations_of_measures_for_one_line in fragment.durations_of_measures:
        for durations_of_measure in durations_of_measures_for_one_line:
            if durations_of_measure not in consistent_patterns:
                score -= 1
    score /= len(fragment.melodic_lines) * (fragment.n_beats // fragment.meter_numerator)
    return score


def find_indices_of_dissonating_events(
        sonority: list[Event], meter_numerator: int
) -> tuple[set[int], set[int]]:
    """
    Find indices of dissonating (i.e., dependent, non-free in terms of strict counterpoint) events.

    :param sonority:
        simultaneously sounding events
    :param meter_numerator:
        numerator in meter signature, i.e., number of reference beats per measure
    :return:
        indices of passing tones or neighbor dissonances and indices of suspended dissonances
    """
    passing_tones_and_neighbors = set()
    suspensions = set()
    sonority_start_time = max(event.start_time for event in sonority)
    pairs = itertools.combinations(sonority, 2)
    for first_event, second_event in pairs:
        if first_event.pitch_class == 'pause' or second_event.pitch_class == 'pause':
            continue
        n_semitones = first_event.position_in_semitones - second_event.position_in_semitones
        is_perfect_fourth_consonant = second_event.line_index != len(sonority) - 1
        interval_type = get_type_of_interval(n_semitones, is_perfect_fourth_consonant)
        if interval_type != IntervalTypes.DISSONANCE:
            continue
        first_event_continues = first_event.start_time < sonority_start_time
        second_event_continues = second_event.start_time < sonority_start_time
        if first_event_continues and second_event_continues:
            continue
        first_event_starts_on_downbeat = first_event.start_time % meter_numerator == 0
        second_event_starts_on_downbeat = second_event.start_time % meter_numerator == 0
        if first_event_continues and second_event_starts_on_downbeat:
            suspensions.add(first_event.line_index)
            continue
        if second_event_continues and first_event_starts_on_downbeat:
            suspensions.add(second_event.line_index)
            continue
        if not first_event_continues:
            passing_tones_and_neighbors.add(first_event.line_index)
        if not second_event_continues:
            passing_tones_and_neighbors.add(second_event.line_index)
    return passing_tones_and_neighbors, suspensions


def find_melodic_interval(
        event: Event, event_index: int, melodic_line: list[Event], shift: int
) -> Optional[int]:
    """
    Find melodic interval between the given event and an adjacent event.

    :param event:
        event (it is assumed that it is not pause)
    :param event_index:
        index of the event in its melodic line
    :param melodic_line:
        melodic line containing the event
    :param shift:
        -1 if interval of arrival to the event is needed or
        1 if interval of departure from the event is needed
    :return:
        size of melodic interval (in semitones)
    """
    try:
        adjacent_event = melodic_line[event_index + shift]
    except IndexError:
        return None
    if adjacent_event.pitch_class == 'pause':
        return None
    n_semitones = event.position_in_semitones - adjacent_event.position_in_semitones
    return n_semitones


def evaluate_dissonances_preparation_and_resolution(
        fragment: Fragment,
        n_semitones_to_pt_and_ngh_preparation_penalty: dict[int, float],
        n_semitones_to_pt_and_ngh_resolution_penalty: dict[int, float],
        n_semitones_to_suspension_resolution_penalty: dict[int, float]
) -> float:
    """
    Evaluate smoothness of dissonances preparation and resolution.

    :param fragment:
        a fragment to be evaluated
    :param n_semitones_to_pt_and_ngh_preparation_penalty:
        mapping from melodic interval size in semitones to a penalty for moving by this interval
        to a dissonance considered to be an analogue of passing tone or neighbor dissonance
    :param n_semitones_to_pt_and_ngh_resolution_penalty:
        mapping from melodic interval size in semitones to a penalty for moving by this interval
        from a dissonance considered to be an analogue of passing tone or neighbor dissonance
    :param n_semitones_to_suspension_resolution_penalty:
        mapping from melodic interval size in semitones to a penalty for moving by this interval
        from a dissonance considered to be an analogue of suspended dissonance
    :return:
        average over all vertical intervals penalty for their preparation and resolution
    """
    score = 0
    n_semitones_to_pt_and_ngh_preparation_penalty[None] = 0
    n_semitones_to_pt_and_ngh_resolution_penalty[None] = 0
    n_semitones_to_suspension_resolution_penalty[None] = 0
    event_indices = [0 for _ in fragment.melodic_lines]
    for sonority in fragment.sonorities:
        zipped = zip(sonority, fragment.melodic_lines, event_indices)
        for event, melodic_line, event_index in zipped:
            if event != melodic_line[event_index]:
                event_indices[event.line_index] += 1
        pt_and_ngh_line_indices, suspension_line_indices = find_indices_of_dissonating_events(
            sonority, fragment.meter_numerator
        )
        for line_index in pt_and_ngh_line_indices:
            event_index = event_indices[line_index]
            melodic_line = fragment.melodic_lines[line_index]
            event = melodic_line[event_index]
            arrival_interval = find_melodic_interval(event, event_index, melodic_line, shift=-1)
            score -= n_semitones_to_pt_and_ngh_preparation_penalty.get(arrival_interval, 1.0)
            departure_interval = find_melodic_interval(event, event_index, melodic_line, shift=1)
            score -= n_semitones_to_pt_and_ngh_resolution_penalty.get(departure_interval, 1.0)
        for line_index in suspension_line_indices:
            event_index = event_indices[line_index]
            melodic_line = fragment.melodic_lines[line_index]
            event = melodic_line[event_index]
            departure_interval = find_melodic_interval(event, event_index, melodic_line, shift=1)
            score -= n_semitones_to_suspension_resolution_penalty.get(departure_interval, 1.0)

    total_n_events = sum(len(melodic_line) for melodic_line in fragment.melodic_lines)
    n_first_events = len(fragment.melodic_lines)
    score /= total_n_events - n_first_events
    return score


def compute_harmonic_stability_of_sonority(
        sonority: list[Event], n_semitones_to_stability: dict[int, float]
) -> float:
    """
    Compute stability of sonority as average stability of intervals forming it.

    :param sonority:
        simultaneously sounding events
    :param n_semitones_to_stability:
        mapping from interval size in semitones to its harmonic stability
    :return:
        average stability of intervals forming the sonority
    """
    stability = 0
    sound_events = [event for event in sonority if event.pitch_class != 'pause']
    if len(sound_events) <= 1:
        return 1.0
    for first, second in itertools.combinations(sound_events, 2):
        interval_in_semitones = abs(first.position_in_semitones - second.position_in_semitones)
        interval_in_semitones %= N_SEMITONES_PER_OCTAVE
        stability += n_semitones_to_stability[interval_in_semitones]
    n_pairs = len(sound_events) * (len(sound_events) - 1) / 2
    stability /= n_pairs
    return stability


def find_sonority_type(
        sonority_start: float, sonority_end: float, regular_positions: list[dict[str, Any]],
        ad_hoc_positions: list[dict[str, Any]], n_beats: int
) -> str:
    """
    Find type of sonority based on its position in time.

    Note that collisions are resolved according to the two following rules:
    1) Ad hoc positions have higher precedence than regular positions;
    2) Precedence amongst either ad hoc positions or regular positions is based on order of
       the positions within the corresponding argument.

    :param sonority_start:
        start time of sonority (in reference beats)
    :param sonority_end:
        end time of sonority (in reference beats)
    :param regular_positions:
        parameters of regular positions (for example, downbeats or relatively strong beats)
    :param ad_hoc_positions:
        parameters of ad hoc positions which appear just once (for example, the beginning of
        the fragment or the 11th reference beat)
    :param n_beats:
        total duration of a fragment (in reference beats)
    :return:
        type of sonority based on its position in time
    """
    for ad_hoc_position in ad_hoc_positions:
        if ad_hoc_position['time'] < 0:
            ad_hoc_position['time'] += n_beats
        if sonority_start <= ad_hoc_position['time'] < sonority_end:
            return ad_hoc_position['name']
    for regular_position in regular_positions:
        denominator = regular_position['denominator']
        ratio = math.floor(sonority_start) // denominator
        processed_start = sonority_start - ratio * denominator
        processed_end = sonority_end - ratio * denominator
        current_time = regular_position['remainder']
        while current_time < processed_end:
            if current_time >= processed_start:
                return regular_position['name']
            current_time += denominator
    return 'default'


def evaluate_harmony_dynamic(
        fragment: Fragment, regular_positions: list[dict[str, Any]],
        ad_hoc_positions: list[dict[str, Any]], ranges: dict[str, tuple[float, float]],
        n_semitones_to_stability: dict[int, float]
) -> float:
    """
    Evaluate dynamic of harmonic stability/tension relatively user-defined desired ranges.

    :param fragment:
        a fragment to be evaluated
    :param regular_positions:
        parameters of regular positions (for example, downbeats or relatively strong beats)
    :param ad_hoc_positions:
        parameters of ad hoc positions which appear just once (for example, the beginning of
        the fragment or the 11th reference beat)
    :param ranges:
        mapping from position type to minimum and maximum desired levels of harmonic stability
    :param n_semitones_to_stability:
        mapping from interval size in semitones to its harmonic stability
    :return:
        average over all sonorities deviation of harmonic stability from its ranges
    """
    score = 0
    for sonority in fragment.sonorities:
        sonority_start = max(event.start_time for event in sonority)
        sonority_end = min(event.start_time + event.duration for event in sonority)
        stability_of_current_sonority = compute_harmonic_stability_of_sonority(
            sonority, n_semitones_to_stability
        )
        sonority_type = find_sonority_type(
            sonority_start, sonority_end, regular_positions, ad_hoc_positions, fragment.n_beats
        )
        min_allowed_value = ranges[sonority_type][0]
        score += min(stability_of_current_sonority - min_allowed_value, 0)
        max_allowed_value = ranges[sonority_type][1]
        score += min(max_allowed_value - stability_of_current_sonority, 0)
    score /= len(fragment.sonorities)
    return score


def evaluate_local_diatonicity(
        fragment: Fragment, depth: int = 2, scale_types: Optional[tuple[str]] = None
) -> float:
    """
    Evaluate presence of diatonic scales at short periods of time.

    If for every short time interval there is a diatonic scale containing all its pitches,
    the fragment might be called pantonal with frequent modulations between diatonic scales.

    :param fragment:
        a fragment to be evaluated
    :param depth:
        duration of a time period as number of successive sonorities
    :param scale_types:
        types of diatonic scales to be tested; this tuple may include the following values:
        'major', 'natural_minor', 'harmonic_minor', 'dorian', 'phrygian', 'lydian', 'mixolydian',
        'locrian', and 'whole_tone'; however, keep in mind that due to relative modes there is no
        need to include all these values; if this argument is not passed, its value is set to
        `('major', 'harmonic_minor', 'whole_tone')` which covers all supported scales
    :return:
        minus one multiplied by average fraction of pitches that are out of the most fitting
        to the current short interval diatonic scale (with averaging over all short intervals)
    """
    score = 0
    scale_types = scale_types or ('major', 'harmonic_minor', 'whole_tone')
    pitch_class_to_diatonic_scales = get_mapping_from_pitch_class_to_diatonic_scales(scale_types)
    nested_pitch_classes = []
    for sonority in fragment.sonorities[:depth - 1]:
        nested_pitch_classes.append([event.pitch_class for event in sonority])
    for sonority in fragment.sonorities[depth - 1:]:
        nested_pitch_classes.append([event.pitch_class for event in sonority])
        pitch_classes = [x for y in nested_pitch_classes for x in y if x != 'pause']
        counter = Counter()
        for pitch_class in pitch_classes:
            counter.update(pitch_class_to_diatonic_scales[pitch_class])
        n_pitch_classes_from_best_scale = counter.most_common(1)[0][1]
        score -= 1 - n_pitch_classes_from_best_scale / len(pitch_classes)
        nested_pitch_classes.pop(0)
    n_periods = len(fragment.sonorities) - depth + 1
    score /= n_periods
    return score


@cache
def get_mapping_from_interval_size_to_character() -> dict[Optional[int], str]:
    """
    Get mapping from interval size (in semitones) to its single-char encoding.

    :return:
        mapping from size (in semitones) of a directed non-compound interval to
        its single-char representation; `None` as key relates to pauses
    """
    intervals = list(range(-N_SEMITONES_PER_OCTAVE, N_SEMITONES_PER_OCTAVE + 1)) + [None]
    result = {interval: letter for interval, letter in zip(intervals, string.ascii_lowercase)}
    return result


def encode_interval(interval: Optional[int]) -> str:
    """
    Encode an interval with a single character.

    :param interval:
        interval size (in semitones)
    :return:
        single-character encoding of the interval
    """
    mapping = get_mapping_from_interval_size_to_character()
    if interval is not None:
        interval = math.copysign(abs(interval) % N_SEMITONES_PER_OCTAVE, interval)
    return mapping[interval]


def encode_line_intervals(melodic_lines: list[list[Event]]) -> list[str]:
    """
    Encode intervals from all melodic lines

    :param melodic_lines:
        melodic lines
    :return:
        strings with encoded intervals for each melodic line
    """
    encoded_lines = []
    for melodic_line in melodic_lines:
        encoded_line = ''
        for previous_event, next_event in zip(melodic_line, melodic_line[1:]):
            if previous_event.pitch_class == 'pause' or next_event.pitch_class == 'pause':
                interval = None
            else:
                interval = next_event.position_in_semitones - previous_event.position_in_semitones
            encoded_line += encode_interval(interval)
        encoded_lines.append(encoded_line)
    return encoded_lines


def generate_elision_patterns(
        motif: tuple[int], original_pattern: str, run: bool = False
) -> list[str]:
    """
    Generate search patterns for modifications of the original motif with one pitch omitted.

    :param motif:
        sequence of directed interval sizes (in semitones) between successive pitches of a motif
    :param original_pattern:
        search pattern (in encoded lines) corresponding to the motif
    :param run:
        flag whether to generate any patterns
    :return:
        generated patterns corresponding to elisions of the motif
    """
    patterns = []
    if not run:
        return patterns
    not_first = f'[^{original_pattern[0]}]'
    first_omission_pattern = not_first + original_pattern[1:]
    patterns.append(first_omission_pattern)
    for i in range(len(motif) - 1):
        motif_with_omission = motif[:i] + (motif[i] + motif[i + 1],) + motif[i + 2:]
        omission_pattern = ''.join([encode_interval(x) for x in motif_with_omission])
        patterns.append(omission_pattern)
    not_last = f'[^{original_pattern[-1]}]'
    last_omission_pattern = original_pattern[:-1] + not_last
    patterns.append(last_omission_pattern)
    return patterns


@cache
def generate_regexps_for_intervallic_motif(
        motif: tuple[int],
        inversion: bool = True, reversion: bool = True,
        elision: bool = False, inverted_elision: bool = False, reverted_elision: bool = False
) -> list['_sre.SRE_Pattern']:
    """
    Generate regular expressions that match intervallic motif and its modifications.

    :param motif:
        sequence of directed interval sizes (in semitones) between successive pitches of a motif
    :param inversion:
        flag whether to include inversion of the original motif
    :param reversion:
        flag whether to include reversion of the original motif
    :param elision:
        flag whether to include modifications of the original motif with one pitch omitted
    :param inverted_elision:
        flag whether to include modifications of the inverted motif with one pitch omitted;
        if `inversion` is set to `False`, this flag affects nothing
    :param reverted_elision:
        flag whether to include modifications of the reverted motif with one pitch omitted;
        if `reversion` is set to `False`, this flag affects nothing
    :return:
        compiled regular expressions
    """
    original_pattern = ''.join([encode_interval(x) for x in motif])
    patterns = [original_pattern]
    patterns.extend(generate_elision_patterns(motif, original_pattern, elision))
    if inversion:
        inverted_motif = tuple(-x for x in motif)
        inverted_pattern = ''.join([encode_interval(x) for x in inverted_motif])
        patterns.append(inverted_pattern)
        patterns.extend(
            generate_elision_patterns(inverted_motif, inverted_pattern, inverted_elision)
        )
    if reversion:
        reverted_motif = motif[::-1]
        reverted_pattern = original_pattern[::-1]
        patterns.append(reverted_pattern)
        patterns.extend(
            generate_elision_patterns(reverted_motif, reverted_pattern, reverted_elision)
        )
    patterns = list(set(patterns))  # In particular, duplicates may occur due to symmetries.
    regexps = [re.compile(pattern) for pattern in patterns]
    return regexps


def evaluate_presence_of_intervallic_motif(
        fragment: Fragment, motif: list[int], min_n_occurrences: list[int],
        inversion: bool = True, reversion: bool = True,
        elision: bool = False, inverted_elision: bool = False, reverted_elision: bool = False
) -> float:
    """
    Evaluate presence of used-defined intervallic motif.

    :param fragment:
        a fragment to be evaluated
    :param motif:
        sequence of directed interval sizes (in semitones) between successive pitches of a motif
    :param min_n_occurrences:
        minimum numbers of motif occurrences for each melodic line
    :param inversion:
        flag whether to include inversion of the original motif
    :param reversion:
        flag whether to include reversion of the original motif
    :param elision:
        flag whether to include modifications of the original motif with one pitch omitted
    :param inverted_elision:
        flag whether to include modifications of the inverted motif with one pitch omitted;
        if `inversion` is set to `False`, this flag affects nothing
    :param reverted_elision:
        flag whether to include modifications of the reverted motif with one pitch omitted;
        if `reversion` is set to `False`, this flag affects nothing
    :return:
        minus one multiplied by relative lack of motif occurrences
    """
    regexps = generate_regexps_for_intervallic_motif(
        tuple(motif), inversion, reversion, elision, inverted_elision, reverted_elision
    )
    encoded_lines = encode_line_intervals(fragment.melodic_lines)
    numerator = 0
    denominator = 0
    for threshold, encoded_line in zip(min_n_occurrences, encoded_lines):
        n_occurrences = 0
        for regexp in regexps:
            n_occurrences += len(re.findall(regexp, encoded_line))
        numerator -= max(threshold - n_occurrences, 0)
        denominator += threshold
    score = numerator / denominator
    return score


def evaluate_rhythmic_homogeneity(fragment: Fragment) -> float:
    """
    Evaluate rhythmic homogeneity between all measures except the last one.

    :param fragment:
        a fragment to be evaluated
    :return:
        a score between minis one and zero depending on rhythmic variation
    """
    score = 0
    for durations_of_measures_for_one_line in fragment.durations_of_measures:
        pairs = itertools.combinations(durations_of_measures_for_one_line[:-1], 2)
        for first_durations, second_durations in pairs:
            first_end_times = list(itertools.accumulate(first_durations))
            second_end_times = list(itertools.accumulate(second_durations))
            avg_n_ends = (len(first_end_times) + len(second_end_times)) / 2
            total_n_ends = len(set(first_end_times + second_end_times))
            score -= (total_n_ends / avg_n_ends - 1)
    n_non_last_measures = (fragment.n_beats // fragment.meter_numerator) - 1
    n_pairs = n_non_last_measures * (n_non_last_measures - 1) / 2
    score /= len(fragment.melodic_lines) * n_pairs
    return score


def evaluate_smoothness_of_voice_leading(
        fragment: Fragment,
        penalty_deduction_per_line: float,
        n_semitones_to_penalty: dict[int, float]
) -> float:
    """
    Evaluate presence of coherent melodic lines that move without large leaps.

    :param fragment:
        a fragment to be evaluated
    :param penalty_deduction_per_line:
        amount of averaged leaps penalty that is deducted for each melodic line
    :param n_semitones_to_penalty:
        mapping from size of melodic interval in semitones to penalty for it
    :return:
        average over melodic lines penalty
    """
    score = 0
    for line in fragment.melodic_lines:
        curr_score = 0
        line_without_pauses = [event for event in line if event.pitch_class != 'pause']
        for first, second in zip(line_without_pauses, line_without_pauses[1:]):
            melodic_interval = abs(first.position_in_semitones - second.position_in_semitones)
            curr_score -= n_semitones_to_penalty.get(melodic_interval, 1.0)
        curr_score /= (len(line_without_pauses) - 1)
        curr_score = min(curr_score + penalty_deduction_per_line, 0)
        score += curr_score
    score /= len(fragment.melodic_lines) * (1 - penalty_deduction_per_line)
    return score


def evaluate_stackability(fragment: Fragment, n_semitones_to_penalty: dict[int, float]) -> float:
    """
    Evaluate seamlessness of repeating a fragment after itself.

    Two copies of a fragment may be stacked together in ostinato or in a period consisting of
    two phrases each of which is the fragment.

    :param fragment:
        a fragment to be evaluated
    :param n_semitones_to_penalty:
        mapping from size of melodic interval (in semitones) between last and first events
        of a melodic line to penalty for it
    :return:
        average over melodic lines penalty
    """
    score = 0
    for line in fragment.melodic_lines:
        first_event = line[0]
        if first_event.pitch_class == 'pause':
            continue
        last_event = line[-1]
        if last_event.pitch_class == 'pause':
            continue
        interval = abs(first_event.position_in_semitones - last_event.position_in_semitones)
        score -= n_semitones_to_penalty.get(interval, 1.0)
    score /= len(fragment.melodic_lines)
    return score


def get_scoring_functions_registry() -> dict[str, Callable]:
    """
    Get mapping from names to corresponding scoring functions.

    :return:
        registry of scoring functions
    """
    registry = {
        'absence_of_aimless_fluctuations': evaluate_absence_of_aimless_fluctuations,
        'absence_of_doubled_pitch_classes': evaluate_absence_of_doubled_pitch_classes,
        'absence_of_simultaneous_skips': evaluate_absence_of_simultaneous_skips,
        'absence_of_voice_crossing': evaluate_absence_of_voice_crossing,
        'cadence_duration': evaluate_cadence_duration,
        'climax_explicity': evaluate_climax_explicity,
        'consistency_of_rhythm_with_meter': evaluate_consistency_of_rhythm_with_meter,
        'dissonances_preparation_and_resolution': evaluate_dissonances_preparation_and_resolution,
        'harmony_dynamic': evaluate_harmony_dynamic,
        'intervallic_motif': evaluate_presence_of_intervallic_motif,
        'local_diatonicity': evaluate_local_diatonicity,
        'rhythmic_homogeneity': evaluate_rhythmic_homogeneity,
        'smoothness_of_voice_leading': evaluate_smoothness_of_voice_leading,
        'stackability': evaluate_stackability,
    }
    return registry


def parse_scoring_sets_registry(params: list[dict[str, Any]]) -> SCORING_SETS_REGISTRY_TYPE:
    """
    Parse mapping from names of scoring sets to scoring sets itself.

    :param params:
        raw parameters of scoring sets obtained from a config
    :return:
        mapping from a name of a scoring set to a list of triples of a scoring function,
        its weight, and its parameters
    """
    scoring_functions_registry = get_scoring_functions_registry()
    scoring_sets_registry = {}
    for scoring_set_params in params:
        scoring_set_name = scoring_set_params['name']
        scoring_fns = []
        for scoring_fn_info in scoring_set_params['scoring_functions']:
            scoring_fn = scoring_functions_registry[scoring_fn_info.pop('name')]
            weights = scoring_fn_info.pop('weights')
            scoring_fns.append((scoring_fn, weights, scoring_fn_info))
        scoring_sets_registry[scoring_set_name] = scoring_fns
    return scoring_sets_registry


def weight_score(unweighted_score: float, weights: dict[float, float]) -> float:
    """
    Transform original score by a piecewise linear function.

    :param unweighted_score:
        original score
    :param weights:
        mapping from a breakpoint to a slope coefficient for an interval from the breakpoint to
        the next-to-the-left breakpoint (or -1 if it is absent)
    :return:
        weighted score
    """
    weighted_score = 0
    breakpoints = list(weights.keys()) + [-1.0]
    for left_point, right_point in zip(breakpoints[1:], breakpoints):
        if right_point <= unweighted_score:
            break
        weighted_score -= (right_point - max(unweighted_score, left_point)) * weights[right_point]
    return weighted_score


def evaluate(
        fragment: Fragment,
        scoring_sets: list[str],
        scoring_sets_registry: SCORING_SETS_REGISTRY_TYPE,
        report: bool = False
) -> tuple[float, str]:
    """
    Evaluate fragment.

    :param fragment:
        fragment to be evaluated
    :param scoring_sets:
        names of scoring sets to be used
    :param scoring_sets_registry:
        mapping from a name of a scoring set to a list of triples of a scoring function,
        its weight, and its parameters
    :param report:
        if it is set to `True`, detailed by functions scores are returned as a second output
    :return:
        weighted sum of scores returned by applicable scoring functions
    """
    score = 0
    report_lines = []
    for scoring_set_name in scoring_sets:
        scoring_set = scoring_sets_registry[scoring_set_name]
        for scoring_fn, weights, params in scoring_set:
            unweighted_score = scoring_fn(fragment, **params)
            curr_score = weight_score(unweighted_score, weights)
            if report:
                fn_name = scoring_fn.__name__.removeprefix('evaluate_')
                report_line = f'{fn_name:>40}: {curr_score}'
                report_lines.append(report_line)
            score += curr_score
    if report:
        report_lines.append(f'Overall score is: {score}')
        report_str = '\n'.join(report_lines)
    else:
        report_str = ""
    return score, report_str



"""
Test `dodecaphony.evaluation` module.

Author: Nikolay Lysenko
"""


from typing import Any, Optional

import pytest

from dodecaphony.evaluation import (
    evaluate,
    evaluate_absence_of_aimless_fluctuations,
    evaluate_absence_of_doubled_pitch_classes,
    evaluate_absence_of_simultaneous_skips,
    evaluate_absence_of_voice_crossing,
    evaluate_cadence_duration,
    evaluate_climax_explicity,
    evaluate_consistency_of_rhythm_with_meter,
    evaluate_dissonances_preparation_and_resolution,
    evaluate_harmony_dynamic,
    evaluate_local_diatonicity,
    evaluate_presence_of_intervallic_motif,
    evaluate_rhythmic_homogeneity,
    evaluate_smoothness_of_voice_leading,
    evaluate_stackability,
    find_indices_of_dissonating_events,
    find_sonority_type,
    parse_scoring_sets_registry,
    weight_score,
)
from dodecaphony.fragment import Event, Fragment, override_calculated_attributes


@pytest.mark.parametrize(
    "fragment, penalties, window_size, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    ['C#', 'D#', 'C', 'E', 'B', 'D', 'F#', 'G#', 'F', 'A', 'G', 'A#'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=12,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[2],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0],
            ),
            # `penalties`
            {2: 1, 6: 0.5},
            # `window_size`
            3,
            # `expected`
            -0.6875
        ),
    ]
)
def test_evaluate_absence_of_aimless_fluctuations(
        fragment: Fragment, penalties: dict[int, float], window_size: int, expected: float
) -> None:
    """Test `evaluate_absence_of_aimless_fluctuations` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_absence_of_aimless_fluctuations(fragment, penalties, window_size)
    assert result == expected


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['B', 'C', 'D#', 'A', 'G', 'A#', 'G#', 'C#', 'E', 'F#', 'D', 'F'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            -0.125
        ),
    ]
)
def test_evaluate_absence_of_doubled_pitch_classes(fragment: Fragment, expected: float) -> None:
    """Test `evaluate_absence_of_doubled_pitch_classes` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_absence_of_doubled_pitch_classes(fragment)
    assert result == expected


@pytest.mark.parametrize(
    "fragment, min_skip_in_semitones, max_skips_share, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['B', 'C', 'D#', 'A', 'G', 'A#', 'G#', 'C#', 'E', 'F#', 'D', 'F'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `min_skip_in_semitones`
            7,
            # `max_skips_share`
            0.1,
            # `expected`
            -2 / 15
        ),
    ]
)
def test_evaluate_absence_of_simultaneous_skips(
        fragment: Fragment, min_skip_in_semitones: int, max_skips_share: float, expected: float
) -> None:
    """Test `evaluate_absence_of_simultaneous_skips` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_absence_of_simultaneous_skips(
        fragment, min_skip_in_semitones, max_skips_share
    )
    assert result == expected


@pytest.mark.parametrize(
    "fragment, n_semitones_to_penalty, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `n_semitones_to_penalty`
            {
                0: 0.5,
                -1: 0.55,
                -2: 0.6,
                -3: 0.65,
                -4: 0.7,
                -5: 0.75,
                -6: 0.8,
                -7: 0.85,
                -8: 0.9,
                -9: 0.95,
                -10: 1
            },
            # `expected`
            -0.55 / 15
        ),
    ]
)
def test_evaluate_absence_of_voice_crossing(
        fragment: Fragment, n_semitones_to_penalty: dict[int, float], expected: float) -> None:
    """Test `evaluate_absence_of_voice_crossing` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_absence_of_voice_crossing(fragment, n_semitones_to_penalty)
    assert round(result, 10) == round(expected, 10)


@pytest.mark.parametrize(
    "fragment, max_duration, last_sonority_weight, last_notes_weight, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `max_duration`
            4,
            # `last_sonority_weight`
            0.9,
            # `last_notes_weight`
            0.1,
            # `expected`
            -0.7375
        ),
    ]
)
def test_evaluate_cadence_duration(
        fragment: Fragment, max_duration: float,
        last_sonority_weight: float, last_notes_weight: float, expected: float
) -> None:
    """Test `evaluate_cadence_duration` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_cadence_duration(
        fragment, max_duration, last_sonority_weight, last_notes_weight
    )
    assert result == expected


@pytest.mark.parametrize(
    "fragment, height_penalties, duplication_penalty, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `height_penalties`
            {
                2: 1.0,
                3: 0.8,
                4: 0.6,
                5: 0.5,
                6: 0.4,
                7: 0.3,
                8: 0.2,
                9: 0.1,
                10: 0.0,
            },
            # `duplication_penalty`
            0.5,
            # `expected`
            -0.2
        )
    ]
)
def test_evaluate_climax_explicity(
        fragment: Fragment, height_penalties: dict[int, float], duplication_penalty: float,
        expected: float
) -> None:
    """Test `evaluate_climax_explicity` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_climax_explicity(fragment, height_penalties, duplication_penalty)
    assert result == expected


@pytest.mark.parametrize(
    "fragment, consistent_patterns, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `consistent_patterns`
            [
                [2.0, 2.0],
                [2.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0],
            ],
            # `expected`
            -0.125
        ),
    ]
)
def test_evaluate_consistency_of_rhythm_with_meter(
        fragment: Fragment, consistent_patterns: list[list[float]], expected: float
) -> None:
    """Test `evaluate_consistency_of_rhythm_with_meter` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_consistency_of_rhythm_with_meter(fragment, consistent_patterns)
    assert result == expected


@pytest.mark.parametrize(
    "fragment, n_semitones_to_pt_and_ngh_preparation_penalty, "
    "n_semitones_to_pt_and_ngh_resolution_penalty, n_semitones_to_suspension_resolution_penalty, "
    "expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 4.0, 4.0, 4.0],
                ],
                sonic_content=[
                    ['G', 'E', 'C', 'G#', 'D#', 'C#', 'A', 'F', 'D', 'A#', 'F#', 'B']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            -0.1 / 9
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 2.0, 2.0, 4.0, 4.0],
                ],
                sonic_content=[
                    ['G', 'E', 'C', 'G#', 'D#', 'pause', 'C#', 'A', 'F', 'D', 'A#', 'F#', 'B']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            -0.01
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 2.0, 2.0, 4.0, 4.0],
                ],
                sonic_content=[
                    ['G', 'E', 'C', 'G#', 'D#', 'C#', 'pause', 'A', 'F', 'D', 'A#', 'F#', 'B']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            -0.01
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 4.0, 4.0, 4.0],
                    [4.0, 4.0, 4.0, 4.0],
                ],
                sonic_content=[
                    ['G#', 'D#', 'C#', 'A#', 'F#', 'B', 'G', 'E', 'C', 'A', 'F', 'D']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            -0.2 / 9
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 4.0, 2.0, 4.0],
                    [2.0, 2.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['E', 'C', 'F', 'D', 'D#', 'B', 'G#', 'C#', 'F#', 'G', 'A', 'A#']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[2],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            -0.1
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 4.0, 1.0, 1.0, 4.0],
                    [2.0, 2.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['E', 'C', 'F', 'D', 'D#', 'pause', 'B', 'G#', 'C#', 'F#', 'G', 'A', 'A#']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[2],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            0
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 2.0, 2.0, 6.0],
                    [2.0, 2.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['E', 'C', 'F', 'D', 'D#', 'B', 'G#', 'C#', 'F#', 'G', 'A', 'A#']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[0, 1],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[2],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_pt_and_ngh_preparation_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_pt_and_ngh_resolution_penalty`
            {
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
            },
            # `n_semitones_to_suspension_resolution_penalty`
            {
                -5: 0.3,
                -4: 0.2,
                -3: 0.1,
                -2: 0.0,
                -1: 0.0,
                0: 0.0,
                1: 0.1,
                2: 0.2,
                3: 0.3,
            },
            # `expected`
            -0.2
        ),
    ]
)
def test_evaluate_dissonances_preparation_and_resolution(
        fragment: Fragment,
        n_semitones_to_pt_and_ngh_preparation_penalty: dict[int, float],
        n_semitones_to_pt_and_ngh_resolution_penalty: dict[int, float],
        n_semitones_to_suspension_resolution_penalty: dict[int, float],
        expected: float
) -> None:
    """Test `evaluate_dissonances_preparation_and_resolution` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_dissonances_preparation_and_resolution(
        fragment,
        n_semitones_to_pt_and_ngh_preparation_penalty,
        n_semitones_to_pt_and_ngh_resolution_penalty,
        n_semitones_to_suspension_resolution_penalty
    )
    assert result == expected


@pytest.mark.parametrize(
    "fragment, regular_positions, ad_hoc_positions, ranges, n_semitones_to_stability, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                ],
                sonic_content=[
                    [
                        'B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F',
                        'pause', 'pause'
                    ]
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `regular_positions`
            [
                {'name': 'downbeat', 'denominator': 4, 'remainder': 0},
                {'name': 'middle', 'denominator': 4, 'remainder': 2},
            ],
            # `ad_hoc_positions`
            [
                {'name': '2nd_beat', 'time': 1.0},
            ],
            # `ranges`
            {
                'downbeat': [0.75, 1.0],
                'middle': [0.5, 0.9],
                '2nd_beat': [0.8, 1.0],
                'default': [0.1, 0.9],
            },
            # `n_semitones_to_stability`
            {
                0: 1.0,
                1: 0.2,
                2: 0.2,
                3: 0.7,
                4: 0.8,
                5: 0.5,
                6: 0.0,
                7: 0.9,
                8: 0.6,
                9: 0.6,
                10: 0.2,
                11: 0.2,
            },
            # `expected`
            -0.1766666667
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                ],
                sonic_content=[
                    [
                        'B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F',
                        'pause', 'pause'
                    ]
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `regular_positions`
            [
                {'name': 'downbeat', 'denominator': 4, 'remainder': 0},
                {'name': 'middle', 'denominator': 4, 'remainder': 2},
            ],
            # `ad_hoc_positions`
            [
                {'name': '2nd_beat', 'time': 1.0},
                {'name': 'ending', 'time': -0.01},
            ],
            # `ranges`
            {
                'downbeat': [0.75, 1.0],
                'middle': [0.5, 0.9],
                '2nd_beat': [0.8, 1.0],
                'ending': [0.0, 0.1],
                'default': [0.1, 0.9],
            },
            # `n_semitones_to_stability`
            {
                0: 1.0,
                1: 0.2,
                2: 0.2,
                3: 0.7,
                4: 0.8,
                5: 0.5,
                6: 0.0,
                7: 0.9,
                8: 0.6,
                9: 0.6,
                10: 0.2,
                11: 0.2,
            },
            # `expected`
            -0.3366666667
        ),
    ]
)
def test_evaluate_harmony_dynamic(
        fragment: Fragment, regular_positions: list[dict[str, Any]],
        ad_hoc_positions: list[dict[str, Any]], ranges: dict[str, tuple[float, float]],
        n_semitones_to_stability: dict[int, float], expected: float
) -> None:
    """Test `evaluate_harmony_dynamic` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_harmony_dynamic(
        fragment, regular_positions, ad_hoc_positions, ranges, n_semitones_to_stability
    )
    assert round(result, 10) == expected


@pytest.mark.parametrize(
    "fragment, depth, scale_types, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `depth`
            2,
            # `scale_types`
            None,
            # `expected`
            -1 / 56
        ),
    ]
)
def test_evaluate_local_diatonicity(
        fragment: Fragment, depth: int, scale_types: Optional[list[str]], expected: float
) -> None:
    """Test `evaluate_local_diatonicity` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_local_diatonicity(fragment, depth, scale_types)
    assert round(result, 10) == round(expected, 10)


@pytest.mark.parametrize(
    "fragment, motif, min_n_occurrences, inversion, reversion, elision, inverted_elision, "
    "reverted_elision, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'pause', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F'],
                    ['B', 'A', 'pause', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `motif`
            [-2, 6, 2, -3],
            # `min_n_occurrences`
            [2, 2],
            # `inversion`
            False,
            # `reversion`
            False,
            # `elision`
            False,
            # `inverted_elision`
            False,
            # `reverted_elision`
            False,
            # `expected`
            -0.75
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'pause', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F'],
                    ['B', 'A', 'pause', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `motif`
            [-2, 6, 2, -3],
            # `min_n_occurrences`
            [2, 2],
            # `inversion`
            True,
            # `reversion`
            True,
            # `elision`
            True,
            # `inverted_elision`
            True,
            # `reverted_elision`
            False,
            # `expected`
            -0.5
        ),
    ]
)
def test_evaluate_presence_of_intervallic_motif(
        fragment: Fragment, motif: list[int], min_n_occurrences: list[int],
        inversion: bool, reversion: bool, elision: bool, inverted_elision: bool,
        reverted_elision: bool, expected: float
) -> None:
    """Test `evaluate_presence_of_intervallic_motif` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_presence_of_intervallic_motif(
        fragment, motif, min_n_occurrences, inversion, reversion,
        elision, inverted_elision, reverted_elision
    )
    assert result == expected


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            -(1 / 9 + (0.6 + 2 / 3 + 1 / 7) / 6)
        ),
    ]
)
def test_evaluate_rhythmic_homogeneity(fragment: Fragment, expected: float) -> None:
    """Test `evaluate_rhythmic_homogeneity` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_rhythmic_homogeneity(fragment)
    assert round(result, 10) == round(expected, 10)


@pytest.mark.parametrize(
    "fragment, penalty_deduction_per_line, n_semitones_to_penalty, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                ],
                sonic_content=[
                    [
                        'B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F',
                        'pause', 'pause'
                    ]
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `penalty_deduction_per_line`
            0.2,
            # `n_semitones_to_penalty`
            {
                0: 0.2,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
                5: 0.3,
                6: 0.4,
                7: 0.5,
                8: 0.6,
                9: 0.7,
                10: 0.8,
                11: 0.9,
                12: 1.0,
            },
            # `expected`
            -1 / 6
        ),
    ]
)
def test_evaluate_smoothness_of_voice_leading(
        fragment: Fragment, penalty_deduction_per_line: float,
        n_semitones_to_penalty: dict[int, float], expected: float
) -> None:
    """Test `evaluate_smoothness_of_voice_leading` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_smoothness_of_voice_leading(
        fragment, penalty_deduction_per_line, n_semitones_to_penalty
    )
    assert round(result, 10) == round(expected, 10)


@pytest.mark.parametrize(
    "fragment, n_semitones_to_penalty, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                ],
                sonic_content=[
                    [
                        'pause', 'B', 'A', 'G', 'C#', 'D#', 'C',
                        'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'
                    ]
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `n_semitones_to_penalty`
            {
                0: 0.2,
                1: 0.0,
                2: 0.0,
                3: 0.1,
                4: 0.2,
                5: 0.3,
                6: 0.4,
                7: 0.5,
                8: 0.6,
                9: 0.7,
                10: 0.8,
                11: 0.9,
                12: 1.0,
            },
            # `expected`
            -2 / 15
        ),
    ]
)
def test_evaluate_stackability(
        fragment: Fragment, n_semitones_to_penalty: dict[int, float], expected: float
) -> None:
    """Test `evaluate_stackability` function."""
    fragment = override_calculated_attributes(fragment)
    result = evaluate_stackability(fragment, n_semitones_to_penalty)
    assert result == expected


@pytest.mark.parametrize(
    "sonority, meter_numerator, expected",
    [
        (
            # `sonority`
            [
                Event(line_index=0, start_time=2.0, duration=4.0, pitch_class='E', position_in_semitones=67),
                Event(line_index=1, start_time=4.0, duration=2.0, pitch_class='D', position_in_semitones=65),
                Event(line_index=2, start_time=4.0, duration=1.0, pitch_class='C', position_in_semitones=63),
            ],
            # `meter_numerator`
            4,
            # `expected`
            ({1, 2}, {0})
        ),
        (
            # `sonority`
            [
                Event(line_index=0, start_time=2.0, duration=4.0, pitch_class='G', position_in_semitones=70),
                Event(line_index=1, start_time=4.0, duration=2.0, pitch_class='E', position_in_semitones=67),
                Event(line_index=2, start_time=4.0, duration=1.0, pitch_class='C', position_in_semitones=63),
            ],
            # `meter_numerator`
            4,
            # `expected`
            (set(), set())
        ),
        (
            # `sonority`
            [
                Event(line_index=0, start_time=2.0, duration=2.0, pitch_class='E', position_in_semitones=67),
                Event(line_index=1, start_time=2.0, duration=2.0, pitch_class='D', position_in_semitones=65),
                Event(line_index=2, start_time=3.0, duration=1.0, pitch_class='C', position_in_semitones=63),
            ],
            # `meter_numerator`
            4,
            # `expected`
            ({2}, set())
        ),
        (
            # `sonority`
            [
                Event(line_index=0, start_time=4.0, duration=2.0, pitch_class='E', position_in_semitones=67),
                Event(line_index=1, start_time=2.0, duration=4.0, pitch_class='D', position_in_semitones=65),
                Event(line_index=2, start_time=4.0, duration=1.0, pitch_class='C', position_in_semitones=63),
            ],
            # `meter_numerator`
            4,
            # `expected`
            (set(), {1})
        ),
        (
            # `sonority`
            [
                Event(line_index=0, start_time=2.0, duration=2.0, pitch_class='pause', position_in_semitones=None),
                Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='D', position_in_semitones=65),
                Event(line_index=2, start_time=2.0, duration=2.0, pitch_class='C', position_in_semitones=63),
            ],
            # `meter_numerator`
            4,
            # `expected`
            ({1}, set())
        ),
        (
            # `sonority`
            [
                Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='E#', position_in_semitones=68),
                Event(line_index=1, start_time=2.0, duration=2.0, pitch_class='D', position_in_semitones=65),
                Event(line_index=2, start_time=2.0, duration=2.0, pitch_class='C', position_in_semitones=63),
            ],
            # `meter_numerator`
            4,
            # `expected`
            ({0}, set())
        ),
    ]
)
def test_find_indices_of_dissonating_events(
        sonority: list[Event], meter_numerator: int, expected: tuple[set[int], set[int]]
) -> None:
    """Test `find_indices_of_dissonating_events` function."""
    result = find_indices_of_dissonating_events(sonority, meter_numerator)
    assert result == expected


@pytest.mark.parametrize(
    "sonority_start, sonority_end, regular_positions, ad_hoc_positions, n_beats, expected",
    [
        (
            # `sonority_start`
            3.5,
            # `sonority_end`
            3.75,
            # `regular_positions`
            [{'name': 't3.5', 'remainder': 3.5, 'denominator': 4}],
            # `ad_hoc_positions`
            [],
            # `n_beats`
            4,
            # `expected`
            't3.5'
        ),
        (
            # `sonority_start`
            3,
            # `sonority_end`
            5,
            # `regular_positions`
            [{'name': 'downbeat', 'remainder': 0, 'denominator': 4}],
            # `ad_hoc_positions`
            [{'name': 'ending', 'time': -0.01}],
            # `n_beats`
            8,
            # `expected`
            'downbeat'
        ),
        (
            # `sonority_start`
            7,
            # `sonority_end`
            8,
            # `regular_positions`
            [{'name': 't3', 'remainder': 3, 'denominator': 4}],
            # `ad_hoc_positions`
            [{'name': 'ending', 'time': -0.01}],
            # `n_beats`
            8,
            # `expected`
            'ending'
        ),
        (
            # `sonority_start`
            3,
            # `sonority_end`
            9,
            # `regular_positions`
            [
                {'name': 'downbeat', 'remainder': 0, 'denominator': 4},
                {'name': 'middle', 'remainder': 2, 'denominator': 4},
            ],
            # `ad_hoc_positions`
            [{'name': 'ending', 'time': -0.01}],
            # `n_beats`
            16,
            # `expected`
            'downbeat'
        ),
        (
            # `sonority_start`
            3,
            # `sonority_end`
            9,
            # `regular_positions`
            [
                {'name': 'middle', 'remainder': 2, 'denominator': 4},
                {'name': 'downbeat', 'remainder': 0, 'denominator': 4},
            ],
            # `ad_hoc_positions`
            [{'name': 'ending', 'time': -0.01}],
            # `n_beats`
            16,
            # `expected`
            'middle'
        ),
    ]
)
def test_find_sonority_type(
        sonority_start: float, sonority_end: float, regular_positions: list[dict[str, Any]],
        ad_hoc_positions: list[dict[str, Any]], n_beats: int, expected: str
) -> None:
    """Test `find_sonority_type` function."""
    result = find_sonority_type(
        sonority_start, sonority_end, regular_positions, ad_hoc_positions, n_beats
    )
    assert result == expected


@pytest.mark.parametrize(
    "fragment, params, scoring_sets, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                    [1.0, 1.0, 1.0, 0.5, 0.5],
                ],
                sonic_content=[
                    [
                        'B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F',
                        'pause', 'pause'
                    ]
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[3],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0],
            ),
            # `params`
            [
                {
                    'name': 'default',
                    'scoring_functions': [
                        {
                            'name': 'smoothness_of_voice_leading',
                            'weights': {0.0: 0.5},
                            'penalty_deduction_per_line': 0.2,
                            'n_semitones_to_penalty': {
                                0: 0.2,
                                1: 0.0,
                                2: 0.0,
                                3: 0.1,
                                4: 0.2,
                                5: 0.3,
                                6: 0.4,
                                7: 0.5,
                                8: 0.6,
                                9: 0.7,
                                10: 0.8,
                                11: 0.9,
                                12: 1.0,
                            },
                        },
                    ],
                }
            ],
            # `scoring_sets`
            ['default'],
            # `expected`
            -1 / 12
        ),
    ]
)
def test_parse_scoring_sets_registry(
        fragment: Fragment, params: list[dict[str, Any]], scoring_sets: list[str], expected: float
) -> None:
    """Test `parse_scoring_sets_registry` function."""
    fragment = override_calculated_attributes(fragment)
    registry = parse_scoring_sets_registry(params)
    result, _ = evaluate(fragment, scoring_sets, registry)
    assert round(result, 10) == round(expected, 10)


@pytest.mark.parametrize(
    "unweighted_score, weights, expected",
    [
        (-0.3, {0.0: 1.0, -0.2: 2.0, -0.5: 3.0}, -0.4),
        (-0.7, {0.0: 1.0, -0.2: 2.0, -0.5: 3.0}, -1.4),
    ]
)
def test_weight_score(
        unweighted_score: float, weights: dict[float, float], expected: float
) -> None:
    """Test `weight_score` function."""
    result = weight_score(unweighted_score, weights)
    assert round(result, 8) == expected



"""
Define data structures representing a fragment of a musical piece.

Author: Nikolay Lysenko
"""


import itertools
import math
import random
from dataclasses import dataclass
from typing import Any, Optional

from sinethesizer.utils.music_theory import get_note_to_position_mapping

from .music_theory import (
    N_SEMITONES_PER_OCTAVE,
    TONE_ROW_LEN,
    get_smallest_intervals_between_pitch_classes,
    invert_tone_row,
    revert_tone_row,
    validate_tone_row
)


NOTE_TO_POSITION_MAPPING = get_note_to_position_mapping()
SMALLEST_INTERVALS_MAPPING = get_smallest_intervals_between_pitch_classes()

SUPPORTED_DURATIONS = [0.25, 0.5, 0.75, 1.0, 1.5, 2.0, 3.0, 4.0, 6.0]
DURATION_SPLITS = {
    6.0: [[4.0, 2.0], [3.0, 3.0]],
    4.0: [[2.0, 2.0], [3.0, 1.0]],
    3.0: [[2.0, 1.0], [1.5, 1.5]],
    2.0: [[1.0, 1.0], [1.5, 0.5]],
    1.5: [[1.0, 0.5], [0.75, 0.75]],
    1.0: [[0.5, 0.5], [0.75, 0.25]],
    0.75: [[0.5, 0.25]],
    0.5: [[0.25, 0.25]],
    0.25: [[0.25]],
}


@dataclass
class Event:
    """An indivisible element of a musical piece."""
    line_index: int
    start_time: float
    duration: float
    pitch_class: Optional[str] = None
    position_in_semitones: Optional[int] = None


@dataclass
class FragmentParams:
    """Initial parameters of a fragment of a musical piece."""
    tone_row: list[str]
    groups: list[dict[str, int]]
    meter_numerator: int
    meter_denominator: int
    n_measures: int
    line_ids: list[int]
    upper_line_highest_note: str
    upper_line_lowest_note: str
    pauses_fraction: float
    temporal_content: Optional[dict[int, dict[str, Any]]] = None
    sonic_content: Optional[dict[int, dict[str, Any]]] = None


@dataclass
class Fragment:
    """A fragment of a musical piece."""
    temporal_content: list[list[float]]
    sonic_content: list[list[str]]
    meter_numerator: int
    meter_denominator: int
    n_beats: int
    line_ids: list[int]
    upper_line_highest_position: int
    upper_line_lowest_position: int
    n_melodic_lines_by_group: list[int]
    n_tone_row_instances_by_group: list[int]
    mutable_temporal_content_indices: list[int]
    mutable_sonic_content_indices: list[int]
    melodic_lines: Optional[list[list[Event]]] = None
    sonorities: Optional[list[list[Event]]] = None
    durations_of_measures: Optional[list[list[list[float]]]] = None

    def __eq__(self, other: Any):
        if not isinstance(other, Fragment):
            return NotImplemented
        return (
            self.temporal_content == other.temporal_content
            and self.sonic_content == other.sonic_content
        )


def validate_initialized_content(params: FragmentParams) -> None:
    """
    Validate parameters related to initialized parts of temporal content and sonic content.

    :param params:
        parameters of a fragment to be created
    :return:
        None
    """
    if params.temporal_content is not None:
        total_duration_in_beats = params.n_measures * params.meter_numerator
        for line_index, line_params in params.temporal_content.items():
            duration_in_beats = sum(line_params['durations'])
            if duration_in_beats != total_duration_in_beats:
                raise ValueError("A line has duration that is not equal to that of the fragment.")
    if params.sonic_content is not None:  # pragma: no branch
        for group_index, group_params in params.sonic_content.items():
            n_sound_events = len([x for x in group_params['pitch_classes'] if x != 'pause'])
            n_tone_row_instances = n_sound_events / TONE_ROW_LEN
            declared_n_instances = params.groups[group_index]['n_tone_row_instances']
            if n_tone_row_instances != declared_n_instances:
                raise ValueError("A group has wrong number of tone row instances.")


def validate(params: FragmentParams) -> None:
    """
    Validate all parameters.

    :param params:
        parameters of a fragment to be created
    :return:
        None
    """
    validate_tone_row(params.tone_row)
    if sum(group['n_melodic_lines'] for group in params.groups) != len(params.line_ids):
        raise ValueError("Number of lines in `groups` is not equal to that in `line_ids`.")
    if params.line_ids != list(set(params.line_ids)):
        raise ValueError("IDs of melodic lines must be unique.")
    if float(params.meter_numerator) not in SUPPORTED_DURATIONS:
        raise ValueError(f"Meter numerator = {params.meter_numerator} is not supported.")
    validate_initialized_content(params)


def split_time_span(n_measures: int, n_events: int, meter_numerator: float) -> list[float]:
    """
    Split time span of one melodic line into time spans of the specified number of events.

    :param n_measures:
        number of measures
    :param n_events:
        number of events
    :param meter_numerator:
        number of reference beats per measure
    :return:
        durations of events (in reference beats)
    """
    if n_events < n_measures:
        raise ValueError("Average duration of an event is longer than semibreve.")
    if n_events > n_measures * meter_numerator / min(SUPPORTED_DURATIONS):
        raise ValueError("The number of events is so high that some of them are too short.")
    durations = [meter_numerator for _ in range(n_measures)]
    current_n_events = n_measures
    index = 0
    while current_n_events < n_events:
        insertion = random.choice(DURATION_SPLITS[durations[index]])
        durations = durations[:index] + insertion + durations[index + 1:]
        index += len(insertion)
        index %= len(durations)
        current_n_events += len(insertion) - 1
    random.shuffle(durations)
    return durations


def calculate_number_of_undefined_events(
        group_index: int,
        temporal_content: list[list[float]],
        sonic_content: dict[int, dict[str, Any]],
        line_indices: list[int],
        n_tone_row_instances: int,
        pauses_fraction: float
) -> int:
    """
    Calculate total number of undefined events in a group of melodic lines sharing the same series.

    :param group_index:
        index of a group of melodic lines
    :param temporal_content:
        partially filled list of lists of event durations (in reference beats)
        for each melodic line
    :param sonic_content:
        mapping from group index to user-defined sequence of its pitch classes (including pauses)
    :param line_indices:
        indices of melodic lines forming the group
    :param n_tone_row_instances:
        number of tone row instances in the group
    :param pauses_fraction:
        fraction of pauses amongst all events from the group
    :return:
        total number of events in the group that are not initially defined by a user
    """
    if group_index in sonic_content:
        n_events = len(sonic_content[group_index]['pitch_classes'])
    else:
        n_sound_events = n_tone_row_instances * TONE_ROW_LEN
        n_events = int(round(n_sound_events / (1 - pauses_fraction)))
    n_events -= sum(len(temporal_content[line_index]) for line_index in line_indices)
    return n_events


def distribute_events_among_lines(n_events: int, n_lines: int) -> list[int]:
    """
    Distribute evenly the specified number of events among the specified number of lines.

    :param n_events:
        number of events
    :param n_lines:
        number of lines
    :return:
        list of numbers of events in a particular line
    """
    results = [n_events // n_lines for _ in range(n_lines)]
    for i in range(n_events - sum(results)):
        results[i] += 1
    return results


def create_initial_temporal_content(params: FragmentParams) -> list[list[float]]:
    """
    Split time span of each melodic line into durations of individual events.

    :param params:
        parameters of a fragment to be created
    :return:
        lists of event durations (in reference beats) for each melodic line
    """
    temporal_content = [
        params.temporal_content.get(line_index, {}).get('durations', [])
        for line_index, line_id in enumerate(params.line_ids)
    ]
    meter_numerator = float(params.meter_numerator)
    group_lines_start_index = 0
    for group_index, group_params in enumerate(params.groups):
        n_lines = group_params['n_melodic_lines']
        line_indices = list(range(group_lines_start_index, group_lines_start_index + n_lines))
        n_undefined_events = calculate_number_of_undefined_events(
            group_index,
            temporal_content,
            params.sonic_content,
            line_indices,
            group_params['n_tone_row_instances'],
            params.pauses_fraction
        )
        indices_of_undefined_lines = [x for x in line_indices if x not in params.temporal_content]
        n_undefined_lines = len(indices_of_undefined_lines)
        n_events_per_line = distribute_events_among_lines(n_undefined_events, n_undefined_lines)
        for line_index, n_events in zip(indices_of_undefined_lines, n_events_per_line):
            durations = split_time_span(params.n_measures, n_events, meter_numerator)
            temporal_content[line_index] = durations
        group_lines_start_index += n_lines
    return temporal_content


def replicate_tone_row(tone_row: list[str], n_instances: int) -> list[str]:
    """
    Create multiple instances of tone row.

    :param tone_row:
        tone row as list of pitch classes (like C or C#, flats are not allowed)
    :param n_instances:
        number of tone row instances
    :return:
        list of pitch classes consisting of instances of tone row (maybe, inverted or reverted)
    """
    pitch_classes = []
    for _ in range(n_instances):
        current_instance = [pitch_class for pitch_class in tone_row]
        if random.choice([True, False]):
            current_instance = invert_tone_row(current_instance)
        if random.choice([True, False]):
            current_instance = revert_tone_row(current_instance)
        pitch_classes.extend(current_instance)
    return pitch_classes


def create_initial_sonic_content(
        params: FragmentParams, temporal_content: list[list[float]]
) -> list[list[str]]:
    """
    Create initial data structure that keeps track of pitch classes.

    :param params:
        parameters of a fragment to be created
    :param temporal_content:
        lists of event durations (in reference beats) for each melodic line
    :return:
        list where for each group of melodic lines sharing the same series sonic content
        of the series is stored (i.e., there is a sequence consisting of pitch classes and pauses)
    """
    sonic_content = []
    group_lines_start_index = 0
    for group_index, group_params in enumerate(params.groups):
        n_lines = group_params['n_melodic_lines']
        if group_index in params.sonic_content:
            sonic_content.append(params.sonic_content[group_index]['pitch_classes'])
            group_lines_start_index += n_lines
            continue

        n_events = 0
        durations = temporal_content[group_lines_start_index:(group_lines_start_index + n_lines)]
        for line_durations in durations:
            n_events += len(line_durations)
        n_sound_events = group_params['n_tone_row_instances'] * TONE_ROW_LEN
        n_pauses = n_events - n_sound_events
        if n_pauses < 0:
            raise ValueError("User-defined temporal content has not enough events.")

        pauses_indices = random.sample(range(n_events), n_pauses)
        series = replicate_tone_row(params.tone_row, group_params['n_tone_row_instances'])
        group_sonic_content = []
        for index in range(n_events):
            group_sonic_content.append('pause' if index in pauses_indices else series.pop(0))
        sonic_content.append(group_sonic_content)

        group_lines_start_index += n_lines
    return sonic_content


def create_grouped_rhythm_only_lines(fragment: Fragment) -> list[list[list[Event]]]:
    """
    Create rhythm-only lines grouped by sharing of the same series.

    :param fragment:
        fragment with non-empty temporal content
    :return:
        list where for each group of melodic lines sharing the same series for each melodic line
        there are events representing temporal (loosely speaking, rhythmic) structure of the line,
        but keeping no information about pitches
    """
    line_index = 0
    rhythm_only_lines = []
    end_indices = list(itertools.accumulate(fragment.n_melodic_lines_by_group))
    start_indices = [0] + end_indices[:-1]
    for start_index, end_index in zip(start_indices, end_indices):
        group_of_lines = []
        for line_durations in fragment.temporal_content[start_index:end_index]:
            start_time = 0
            line = []
            for duration in line_durations:
                line.append(Event(line_index, start_time, duration))
                start_time += duration
            group_of_lines.append(line)
            line_index += 1
        rhythm_only_lines.append(group_of_lines)
    return rhythm_only_lines


def distribute_pitch_classes(fragment: Fragment) -> list[list[Event]]:
    """
    Distribute pitch classes among melodic lines.

    :param fragment:
        fragment with non-empty temporal content and sonic content
    :return:
        melodic lines with pitch classes, but without exact pitches;
        these lines are derived from temporal content and sonic content of the fragment
    """
    melodic_lines = [[] for _ in fragment.line_ids]
    grouped_rhythm_only_lines = create_grouped_rhythm_only_lines(fragment)
    zipped = zip(grouped_rhythm_only_lines, fragment.sonic_content)
    for group_of_rhythm_only_lines, group_sonic_content in zipped:
        timeline = [
            event
            for rhythm_only_line in group_of_rhythm_only_lines
            for event in rhythm_only_line
        ]
        timeline = sorted(timeline, key=lambda event: (event.start_time, event.line_index))
        for event, pitch_class in zip(timeline, group_sonic_content):
            new_event = Event(event.line_index, event.start_time, event.duration, pitch_class)
            melodic_lines[event.line_index].append(new_event)
    return melodic_lines


def find_sonorities(melodic_lines: list[list[Event]]) -> list[list[Event]]:
    """
    Find simultaneously sounding events.

    :param melodic_lines:
        melodic lines
    :return:
        list of simultaneously sounding events
    """
    timeline = [event for melodic_line in melodic_lines for event in melodic_line]
    timeline = sorted(timeline, key=lambda event: (event.start_time, event.line_index))
    indices = [-1 for _ in melodic_lines]
    current_times = [0 for _ in melodic_lines]
    previous_passed_time = 0
    sonorities = []
    for event in timeline:
        indices[event.line_index] += 1
        current_times[event.line_index] += event.duration
        passed_time = min(current_times)
        if passed_time > previous_passed_time:
            sonorities.append([
                melodic_line[index] for melodic_line, index in zip(melodic_lines, indices)
            ])
        previous_passed_time = passed_time
    return sonorities


def transpose_up(position: int, min_position: int) -> int:
    """
    Transpose pitch up by minimum number of octaves that is enough to place it above threshold.

    :param position:
        position of pitch (in semitones)
    :param min_position:
        minimum high enough position (in semitones)
    :return:
        transposed pitch
    """
    shortage = max(min_position - position, 0)
    position += int(math.ceil(shortage / N_SEMITONES_PER_OCTAVE)) * N_SEMITONES_PER_OCTAVE
    return position


def transpose_down(position: int, max_position: int) -> int:
    """
    Transpose pitch down by minimum number of octaves that is enough to place it below threshold.

    :param position:
        position of pitch (in semitones)
    :param max_position:
        maximum low enough position (in semitones)
    :return:
        transposed pitch
    """
    surplus = max(position - max_position, 0)
    position -= int(math.ceil(surplus / N_SEMITONES_PER_OCTAVE)) * N_SEMITONES_PER_OCTAVE
    return position


def set_pitches_of_upper_line(fragment: Fragment) -> Fragment:
    """
    Set exact pitches of events from upper line.

    :param fragment:
        fragment with `melodic_lines` attribute where pitch classes are set
    :return:
        fragment with `melodic_lines` attribute where exact pitch of upper line events are set
    """
    upper_line = fragment.melodic_lines[0]
    for index, event in enumerate(upper_line):  # pragma: no branch
        if event.pitch_class != 'pause':
            break
    low_octave = '1'
    position = NOTE_TO_POSITION_MAPPING[event.pitch_class + low_octave]
    position = transpose_up(position, fragment.upper_line_lowest_position)
    event.position_in_semitones = position
    previous_event_pitch_class = event.pitch_class
    for event in upper_line[index + 1:]:
        if event.pitch_class == 'pause':
            continue
        interval = SMALLEST_INTERVALS_MAPPING[(previous_event_pitch_class, event.pitch_class)]
        position += interval
        position = transpose_up(position, fragment.upper_line_lowest_position)
        position = transpose_down(position, fragment.upper_line_highest_position)
        event.position_in_semitones = position
        previous_event_pitch_class = event.pitch_class
    return fragment


def set_pitches_of_lower_lines(
        fragment: Fragment, max_interval: int = 16, default_shift: int = 7
) -> Fragment:
    """
    Set exact pitches of events from all melodic lines except the upper one.

    :param fragment:
        fragment with `melodic_lines` attribute where pitch classes are set and exact pitches
        are set in the upper line
    :param max_interval:
        maximum interval (in semitones) between two simultaneously sounding events from adjacent
        melodic lines
    :param default_shift:
        downward shift (in semitones) of upper threshold when upper event is pause
    :return:
        fragment where pitches are set for all events
    """
    upper_first_event = fragment.melodic_lines[0][0]
    threshold = upper_first_event.position_in_semitones or fragment.upper_line_lowest_position
    previous_positions = [threshold] + [None for _ in range(len(fragment.melodic_lines) - 1)]
    previous_pitch_classes = [None for _ in range(len(fragment.melodic_lines))]
    for sonority in fragment.sonorities:
        sonority_start = max(event.start_time for event in sonority)
        threshold = sonority[0].position_in_semitones or previous_positions[0]
        previous_positions[0] = threshold
        threshold -= 1  # It is inclusive, so subtraction prevents lines from overlapping.
        for event in sonority[1:]:
            if event.pitch_class == 'pause':
                threshold -= default_shift
                continue
            if event.start_time < sonority_start:
                threshold = event.position_in_semitones - 1
                continue
            previous_pitch_class = previous_pitch_classes[event.line_index]
            if previous_pitch_class is None:
                high_octave = '7'
                position = NOTE_TO_POSITION_MAPPING[event.pitch_class + high_octave]
            else:
                interval = SMALLEST_INTERVALS_MAPPING[(previous_pitch_class, event.pitch_class)]
                position = previous_positions[event.line_index] + interval
            position = transpose_down(position, threshold)
            position = transpose_up(position, 0)
            if threshold - position > max_interval:
                position += N_SEMITONES_PER_OCTAVE
            event.position_in_semitones = position
            previous_positions[event.line_index] = position
            previous_pitch_classes[event.line_index] = event.pitch_class
            threshold = position - 1
    return fragment


def calculate_durations_of_measures(fragment: Fragment) -> list[list[list[float]]]:
    """
    Calculate durations of measures.

    :param fragment:
        fragment with `melodic_lines` attribute
    :return:
        list where for each melodic line there is a list where each measure is represented as list
        of durations of events forming it; suspended from the previous measure event is partially
        included and suspended to the next measure event is fully included
    """
    durations_of_measures = []
    for melodic_line in fragment.melodic_lines:
        rhythm_data = []
        for event in melodic_line:
            record = {
                'measure_id': event.start_time // fragment.meter_numerator,
                'time_since_measure_start': event.start_time % fragment.meter_numerator,
                'duration': event.duration
            }
            rhythm_data.append(record)

        durations_of_measures_for_one_line = []
        for measure_id, values in itertools.groupby(rhythm_data, lambda x: x['measure_id']):
            first_value = values.__next__()
            if first_value['time_since_measure_start'] != 0:
                durations = [first_value['time_since_measure_start'], first_value['duration']]
            else:
                durations = [first_value['duration']]
            for value in values:
                durations.append(value['duration'])
            durations_of_measures_for_one_line.append(durations)
        durations_of_measures.append(durations_of_measures_for_one_line)
    return durations_of_measures


def override_calculated_attributes(fragment: Fragment) -> Fragment:
    """
    Override calculated attributes with values computed based on temporal and sonic content.

    :param fragment:
        fragment to be updated
    :return:
        updated fragment
    """
    fragment.melodic_lines = distribute_pitch_classes(fragment)
    fragment.sonorities = find_sonorities(fragment.melodic_lines)
    fragment = set_pitches_of_upper_line(fragment)
    fragment = set_pitches_of_lower_lines(fragment)
    fragment.durations_of_measures = calculate_durations_of_measures(fragment)
    return fragment


def find_mutable_temporal_content_indices(params: FragmentParams) -> list[int]:
    """
    Find indices of melodic lines such that their temporal content can be modified.

    :param params:
        parameters of a fragment to be created
    :return:
        indices of melodic lines such that their temporal content can be modified
    """
    results = []
    for line_index, line_id in enumerate(params.line_ids):
        if not params.temporal_content.get(line_index, {}).get('immutable', False):
            results.append(line_index)
    return results


def find_mutable_sonic_content_indices(params: FragmentParams) -> list[int]:
    """
    Find indices of groups such that their sonic content can be modified.

    :param params:
        parameters of a fragment to be created
    :return:
        indices of groups such that their sonic content can be modified
    """
    results = []
    for group_index, group_params in enumerate(params.groups):
        if not params.sonic_content.get(group_index, {}).get('immutable', False):
            results.append(group_index)
    return results


def initialize_fragment(params: FragmentParams) -> Fragment:
    """
    Create fragment by its parameters.

    :param params:
        parameters of a fragment to be created
    :return:
        fragment
    """
    params.temporal_content = params.temporal_content or {}
    params.sonic_content = params.sonic_content or {}
    validate(params)
    temporal_content = create_initial_temporal_content(params)
    sonic_content = create_initial_sonic_content(params, temporal_content)
    fragment = Fragment(
        temporal_content,
        sonic_content,
        params.meter_numerator,
        params.meter_denominator,
        params.n_measures * params.meter_numerator,
        params.line_ids,
        NOTE_TO_POSITION_MAPPING[params.upper_line_highest_note],
        NOTE_TO_POSITION_MAPPING[params.upper_line_lowest_note],
        [group['n_melodic_lines'] for group in params.groups],
        [group['n_tone_row_instances'] for group in params.groups],
        find_mutable_temporal_content_indices(params),
        find_mutable_sonic_content_indices(params),
    )
    fragment = override_calculated_attributes(fragment)
    return fragment
